@@ -87,6 +87,9 @@ class CommandLineInterface(object):
         #: EditingMode.VI or EditingMode.EMACS
         self.editing_mode = application.editing_mode
 
+        #: Quoted insert. This flag is set if we go into quoted insert mode.
+        self.quoted_insert = False
+
         #: Vi state. (For Vi key bindings.)
         self.vi_state = ViState()
 
@@ -145,6 +148,10 @@ class CommandLineInterface(object):
     def clipboard(self):
         return self.application.clipboard
 
+    @property
+    def pre_run_callables(self):
+        return self.application.pre_run_callables
+
     def add_buffer(self, name, buffer, focus=False):
         """
         Insert a new buffer.
@@ -268,7 +275,10 @@ class CommandLineInterface(object):
         """
         Reset everything, for reading the next input.
 
-        :param reset_current_buffer: If True, also reset the focussed buffer.
+        :param reset_current_buffer: XXX: not used anymore. The reason for
+            having this option in the past was when this CommandLineInterface
+            is run multiple times, that we could reset the buffer content from
+            the previous run. This is now handled in the AcceptAction.
         """
         # Notice that we don't reset the buffers. (This happens just before
         # returning, and when we have multiple buffers, we clearly want the
@@ -285,9 +295,6 @@ class CommandLineInterface(object):
         self.layout.reset()
         self.vi_state.reset()
 
-        if reset_current_buffer:
-            self.current_buffer.reset()
-
         # Search new search state. (Does also remember what has to be
         # highlighted.)
         self.search_state = SearchState(ignore_case=Condition(lambda: self.is_ignoring_case))
@@ -365,12 +372,27 @@ class CommandLineInterface(object):
         self.renderer.request_absolute_cursor_position()
         self._redraw()
 
+    def _load_next_buffer_indexes(self):
+        for buff, index in self._next_buffer_indexes.items():
+            if buff in self.buffers:
+                self.buffers[buff].working_index = index
+
+    def _pre_run(self, pre_run=None):
+        " Called during `run`. "
+        if pre_run:
+            pre_run()
+
+        # Process registered "pre_run_callables" and clear list.
+        for c in self.pre_run_callables:
+            c()
+        del self.pre_run_callables[:]
+
     def run(self, reset_current_buffer=False, pre_run=None):
         """
         Read input from the command line.
         This runs the eventloop until a return value has been set.
 
-        :param reset_current_buffer: Reset content of current buffer.
+        :param reset_current_buffer: XXX: Not used anymore.
         :param pre_run: Callable that is called right after the reset has taken
             place. This allows custom initialisation.
         """
@@ -380,11 +402,10 @@ class CommandLineInterface(object):
             self._is_running = True
 
             self.on_start.fire()
-            self.reset(reset_current_buffer=reset_current_buffer)
+            self.reset()
 
             # Call pre_run.
-            if pre_run:
-                pre_run()
+            self._pre_run(pre_run)
 
             # Run eventloop in raw mode.
             with self.input.raw_mode():
@@ -433,11 +454,10 @@ class CommandLineInterface(object):
                     self._is_running = True
 
                     self.on_start.fire()
-                    self.reset(reset_current_buffer=reset_current_buffer)
+                    self.reset()
 
                     # Call pre_run.
-                    if pre_run:
-                        pre_run()
+                    self._pre_run(pre_run)
 
                     with self.input.raw_mode():
                         self.renderer.request_absolute_cursor_position()
@@ -591,7 +611,7 @@ class CommandLineInterface(object):
         if self.eventloop:
             self.eventloop.stop()
 
-    def run_in_terminal(self, func, render_cli_done=False):
+    def run_in_terminal(self, func, render_cli_done=False, cooked_mode=True):
         """
         Run function on the terminal above the prompt.
 
@@ -604,6 +624,8 @@ class CommandLineInterface(object):
         :param render_cli_done: When True, render the interface in the
                 'Done' state first, then execute the function. If False,
                 erase the interface first.
+        :param cooked_mode: When True (the default), switch the input to
+                cooked mode while executing the function.
 
         :returns: the result of `func`.
         """
@@ -617,10 +639,12 @@ class CommandLineInterface(object):
         self._return_value = None
 
         # Run system command.
-        with self.input.cooked_mode():
+        if cooked_mode:
+            with self.input.cooked_mode():
+                result = func()
+        else:
             result = func()
 
-
         # Redraw interface again.
         self.renderer.reset()
         self.renderer.request_absolute_cursor_position()
